from utils import *
from sqlalchemy import *

from dbSchemaConsts import *
from sqlManager import *
import globals

class TableDataProvider:
    """
    Parent class for all table data providers
    """

    bindParams = None

    def __init__(self, orderby = 0, asc = True):
        self.orderby = orderby
        self.asc = asc
        self._validate()
        self._load()

    """
    Template method, childs have to overload methods that they need
    """
    def _load(self):
        self._buildQuery()
        self._validate()
        if (self.bindParams == None):
            sqlResult = executeQuery(self.query)
        else:
            sqlResult = executeQuery(self.query, self.bindParams)
        self.data = self._sqlResultToData(sqlResult)
        self.qtime = sqlResult.time
        sqlResult.close()

    def _validate(self):
        if (self.orderby < 0 or self.orderby >= len(self.cols)):
            globals.logger.log("Bad orderby col number provided for " + str(self))
            self.orderby = 0

    def _sqlResultToData(self, sqlResult):
        data = []
        for line in sqlResult: data.append(line)
        return data

class StatsDataProvider:

    def __init__(self, title, tableName):
        self.title = title
        self.tableName = tableName
        self.cols = (
            ["New jobs", lambda x: formJobTableLink(x, tableName, JOB_NEW)],
            ["Used jobs", lambda x: formJobTableLink(x, tableName, JOB_USED)],
            ["Successful jobs", lambda x: formJobTableLink(x, tableName, JOB_SUCCESS)],
            ["Failed jobs", lambda x: formJobTableLink(x, tableName, JOB_FAILURE)],
            ["Abandoned jobs", lambda x: formJobTableLink(x, tableName, JOB_ABANDONED)],
        )
        self.jobTable = Table(self.tableName, globals.meta, autoload = True)
        self.qtime = 0;
        self.data = []
        self.data.append([])
        for i in range (JOB_NEW, JOB_ABANDONED + 1):
            query = select([func.count(self.jobTable.c.job_status)], self.jobTable.c.job_status == i)
            sqlResult = executeQuery(query)
            self.data[0].append(sqlResult.fetchone()[0])
            self.qtime += sqlResult.time
            sqlResult.close()

class MergeJobsDataProvider(TableDataProvider):

    cols = [
        ["Job ID", None],
        ["Run ID", formRunLink],
        ["Running time", deltaTime],
        ["Repacked", None],
        ["Dataset", None]
    ]

    def __init__(self, orderby = 0, asc = True, status = None):
        self.status = status
        TableDataProvider.__init__(self, orderby, asc)
        s = ""
        if JOB_STATUS.has_key(status): s = JOB_STATUS[status]
        if (self.status != None):
            self.title = "Merge " + s + " jobs"
            self.cols[2] = ["Definition time", formatTime]
        else:
            self.title = str(len(self.data)) + " merge jobs running"
            self.cols[2] = ["Running time", deltaTime]

    def _buildQuery(self):
        jobTable = Table('merge_job_def', globals.meta, autoload = True)
        statusTable = Table('job_status', globals.meta, autoload = True)
        assocTable = Table('merge_job_repack_assoc', globals.meta, autoload = True)
        repackTable = Table('repacked', globals.meta, autoload = True)
        datasetTable = Table('primary_dataset', globals.meta, autoload = True)

        c = (
            jobTable.c.job_id,
            repackTable.c.run_id,
            jobTable.c.definition_time,
            func.count(assocTable.c.repacked_id),
            datasetTable.c.name,
        )

        if (self.status != None):
            cond = jobTable.c.job_status == self.status
        else:
            cond = or_(
                jobTable.c.job_status == JOB_USED,
                jobTable.c.job_status == JOB_NEW,
            )

        if self.asc: o = c[self.orderby].asc()
        else: o = c[self.orderby].desc()

        self.query = select(
            c,
            cond,
            from_obj = [
                jobTable.
                #outerjoin(statusTable).
                outerjoin(assocTable).
                outerjoin(repackTable, repackTable.c.repacked_id == assocTable.c.repacked_id).
                outerjoin(datasetTable, datasetTable.c.dataset_id == repackTable.c.dataset_id)
            ]
            ).group_by(
                jobTable.c.job_id,
                jobTable.c.definition_time,
                datasetTable.c.name,
                repackTable.c.run_id,
            ).order_by(o)

class RecoJobsDataProvider(TableDataProvider):

    cols = [
        ["Job ID", None],
        ["Run ID", formRunLink],
        ["Running time", deltaTime],
        ["Repacked", None],
        ["Dataset", None]
    ]

    def __init__(self, orderby = 0, asc = True, status = None):
        self.status = status
        s = ""
        if JOB_STATUS.has_key(status): s = JOB_STATUS[status]
        TableDataProvider.__init__(self, orderby, asc)
        if (self.status != None):
            self.title = "Reco " + s + " jobs"
            self.cols[2] = ["Definition time", formatTime]
        else:
            self.title = str(len(self.data)) + " reco jobs running"
            self.cols[2] = ["Running time", deltaTime]

    def _buildQuery(self):
        jobTable = Table('promptreco_job_def', globals.meta, autoload = True)
        statusTable = Table('job_status', globals.meta, autoload = True)
        assocTable = Table('promptreco_job_repack_assoc', globals.meta, autoload = True)
        repackTable = Table('repacked', globals.meta, autoload = True)
        datasetTable = Table('primary_dataset', globals.meta, autoload = True)

        c = (
             jobTable.c.job_id,
             repackTable.c.run_id,
             jobTable.c.definition_time,
             func.count(assocTable.c.repacked_id),
             datasetTable.c.name,
        )

        if self.asc: o = c[self.orderby].asc()
        else: o = c[self.orderby].desc()

        if (self.status != None):
            cond = jobTable.c.job_status == self.status
        else:
            cond = or_(
                jobTable.c.job_status == JOB_USED,
                jobTable.c.job_status == JOB_NEW,
            )

        self.query = select(
            c,
            cond,
            from_obj = [
                jobTable.
                #outerjoin(statusTable).
                outerjoin(assocTable).
                outerjoin(repackTable, repackTable.c.repacked_id == assocTable.c.repacked_id).
                outerjoin(datasetTable, datasetTable.c.dataset_id == repackTable.c.dataset_id)
            ]
            ).group_by(
                jobTable.c.job_id,
                jobTable.c.definition_time,
                datasetTable.c.name,
                repackTable.c.run_id,
            ).order_by(o)

class RepackJobsDataProvider(TableDataProvider):

    cols = [
        ["Job ID", None],
        ["Run ID", formRunLink],
        ["Running time", deltaTime],
        ["Streamers", None],
        ["Stream", None]
        # ["Dataset", None]
    ]

    def __init__(self, orderby = 0, asc = True, status = None):
        self.status = status
        s = ""
        if JOB_STATUS.has_key(status): s = JOB_STATUS[status]
        TableDataProvider.__init__(self, orderby, asc)
        if (self.status != None):
            self.title = "Repacker " + s + " jobs"
            self.cols[2] = ["Definition time", formatTime]
        else:
            self.title = str(len(self.data)) + " repack jobs running"
            self.cols[2] = ["Running time", deltaTime]

    def _buildQuery(self):
        jobTable = Table('repack_job_def', globals.meta, autoload = True)
        assocTable = Table('job_dataset_streamer_assoc', globals.meta, autoload = True)
        statusTable = Table('job_status', globals.meta, autoload = True)
        datasetTable = Table('primary_dataset', globals.meta, autoload = True)
        streamerTable = Table('streamer', globals.meta, autoload = True)

        c = (
             jobTable.c.job_id,
             streamerTable.c.run_id,
             jobTable.c.definition_time,
             func.count(assocTable.c.streamer_id),
             streamerTable.c.streamname,
             # datasetTable.c.name,
        )

        if (self.status != None):
            cond = jobTable.c.job_status == self.status
        else:
            cond = or_(
                jobTable.c.job_status == JOB_USED,
                jobTable.c.job_status == JOB_NEW,
            )

        if self.asc: o = c[self.orderby].asc()
        else: o = c[self.orderby].desc()

        self.query = select(
            c,
            cond,
            from_obj = [
                jobTable.
                outerjoin(assocTable).
                # outerjoin(datasetTable, datasetTable.c.dataset_id == assocTable.c.dataset_id).
                outerjoin(streamerTable, streamerTable.c.streamer_id == assocTable.c.streamer_id)
            ]
            ).group_by(
                jobTable.c.job_id,
                jobTable.c.definition_time,
                # datasetTable.c.name,
                streamerTable.c.run_id,
                streamerTable.c.streamname,
            ).order_by(o)
        # globals.logger.log(self.query)

class RunRepackDataProvider(TableDataProvider):

    cols = (
        ["Id", None],
        ["Filesize", humanFilesize],
        ["Events", None],
        ["Status", None],
        ["Export status", None],
        ["Dataset", None],
        ["Lfn", lambda x: textToImageAlt(x)]
    )

    def __init__(self, runid, orderby = 0, asc = True):
        self.runid = runid
        TableDataProvider.__init__(self, orderby, asc)
        self.title = str(len(self.data)) + " repacked files"

    def _buildQuery(self):
        repackedTable = Table('repacked', globals.meta, autoload = True)
        rStatusTable = Table('repacked_status', globals.meta, autoload = True)
        eStatusTable = Table('export_status', globals.meta, autoload = True)
        datasetTable = Table('primary_dataset', globals.meta, autoload = True)

        c = (
         repackedTable.c.repacked_id,
         repackedTable.c.filesize,
         repackedTable.c.events,
         rStatusTable.c.status,
         eStatusTable.c.status,
         datasetTable.c.name,
         repackedTable.c.lfn
        )

        if self.asc: o = c[self.orderby].asc()
        else: o = c[self.orderby].desc()

        self.query = select(
            c,
            repackedTable.c.run_id == self.runid,
            from_obj = [repackedTable.outerjoin(rStatusTable).outerjoin(eStatusTable, repackedTable.c.export_status == eStatusTable.c.id).outerjoin(datasetTable)]
            ).order_by(o)

class RunRecoDataProvider(TableDataProvider):

    cols = (
        ["Id", None],
        ["Filesize", humanFilesize],
        ["Events", None],
        ["Status", None],
        ["Export status", None],
        ["Dataset", None],
        ["Lfn", lambda x: textToImageAlt(x)]
    )

    def __init__(self, runid, orderby = 0, asc = True):
        self.runid = runid
        TableDataProvider.__init__(self, orderby, asc)
        self.title = str(len(self.data)) + " reconstructed files"

    def _buildQuery(self):
        recoTable = Table('reconstructed', globals.meta, autoload = True)
        rStatusTable = Table('reconstructed_status', globals.meta, autoload = True)
        eStatusTable = Table('export_status', globals.meta, autoload = True)
        datasetTable = Table('primary_dataset', globals.meta, autoload = True)

        c = (
            recoTable.c.reconstructed_id,
            recoTable.c.filesize,
            recoTable.c.events,
            rStatusTable.c.status,
            eStatusTable.c.status,
            datasetTable.c.name,
            recoTable.c.lfn
        )

        if self.asc: o = c[self.orderby].asc()
        else: o = c[self.orderby].desc()

        self.query = select(
            c,
            recoTable.c.run_id == self.runid,
            from_obj = [
                recoTable.
                outerjoin(rStatusTable).
                outerjoin(eStatusTable, recoTable.c.export_status == eStatusTable.c.id).
                outerjoin(datasetTable, datasetTable.c.dataset_id == recoTable.c.dataset_id)
            ]
            ).order_by(o)

class RunStreamerDataProvider(TableDataProvider):

    cols = (
        ["Id", None],
        ["Lumi Id", None],
        ["Filesize", humanFilesize],
        ["Events", None],
        ["Name", None],
        ["E", None],
        ["S", None],
        ["Lfn", lambda x: textToImageAlt(x)]
    )

    def __init__(self, runid, orderby = 0, asc = True):
        self.runid = runid
        TableDataProvider.__init__(self, orderby, asc)
        self.title = str(len(self.data)) + " streamer files"

    def _buildQuery(self):
        streamerTable = Table('streamer', globals.meta, autoload = True)

        c = (
            streamerTable.c.streamer_id,
            streamerTable.c.lumi_id,
            streamerTable.c.filesize,
            streamerTable.c.events,
            streamerTable.c.streamname,
            streamerTable.c.exportable,
            streamerTable.c.splitable,
            streamerTable.c.lfn
        )

        if self.asc: o = c[self.orderby].asc()
        else: o = c[self.orderby].desc()

        self.query = select(
            c,
            streamerTable.c.run_id == self.runid
        ).order_by(o)

class RunsDataProvider(TableDataProvider):

    cols = (
        ["Run ID", formRunLink],
        ["Start time", formatTime],
        ["End time", formatTime],
        ["Process", None],
        ["Version", None],
        ["Global tag", None],
        ["Status", None],
        ["Streamers", None],
        ["Repacked", None],
        ["Reconstructed", None],
        # ["Num repacked", None],
        # ["Num reconstructed", None],
    )

    def __init__(self, numruns = 5, orderby = 0, asc = True):
        self.numruns = numruns;
        self.title = "Top " + str(self.numruns) + " runs"
        TableDataProvider.__init__(self, orderby, asc)

    def _buildQuery(self):
#        runTable = Table('run', globals.meta, autoload = True)
#        runStatusTable = Table('run_status', globals.meta, autoload = True)
#        streamerTable = Table('streamer', globals.meta, autoload = True)

#        c = (
#            runTable.c.run_id,
#            runTable.c.start_time,
#            runTable.c.end_time,
#            runTable.c.process,
#            runTable.c.version,
#            runStatusTable.c.status,
#            "count_1",
#            # select(func.count(streamerTable.c.run_id), streamerTable.c.run_id == runTable.c.run_id)
#            # func.count(streamerTable.c.run_id),
#            # func.count(reconstructedTable.c.run_id)
#        )

#        if self.asc: o = c[self.orderby].asc()
#        else: o = c[self.orderby].desc()
#
#        self.query = select(
#            c,
#            from_obj = [
#                select(
#                    [func.count(streamerTable.c.streamer_id)],
#                    # streamerTable.c.run_id == runTable.c.run_id
#                    ),
#                runTable.
#                outerjoin(runStatusTable),
#                # outerjoin(streamerTable, runTable.c.run_id == streamerTable.c.run_id),
#            ]
#            ).order_by(o).limit(self.numruns)
#            #group_by(
#            #    runTable.c.run_id,
#            #    runTable.c.start_time,
#            #    runTable.c.end_time,
#            #    runTable.c.process,
#            #    runTable.c.version,
#            #    runStatusTable.c.status,
#            #).
#            # order_by(o).limit(self.numruns)

        if self.asc: order = "ASC"
        else: order = "DESC"

        # HACK: query time optimisation
        if self.orderby <= 6:
            self.query = """
            SELECT sub.run_id, sub.start_time, sub.end_time, sub.process, sub.version, sub.tag, sub.status,
            (SELECT COUNT(s.run_id) FROM streamer s WHERE s.run_id = sub.run_id) numStreamers,
            (SELECT COUNT(re.run_id) FROM repacked re WHERE re.run_id = sub.run_id) numRepacked,
            (SELECT COUNT(rec.run_id) FROM reconstructed rec WHERE rec.run_id = sub.run_id) numReconstructed
            FROM(
                SELECT
                r.run_id run_id,
                r.start_time start_time,
                r.end_time end_time,
                r.process process,
                r.version version,
                r.global_tag tag,
                rs.status status
                FROM run r, run_status rs
                WHERE rs.id = r.run_status
                ORDER BY """ + str(int(self.orderby) + 1) + " " + order + """
            ) sub WHERE rownum <= :numruns"""
        else:
            # slow query
            self.query = """
            SELECT *
            FROM(
                SELECT
                r.run_id run_id,
                r.start_time start_time,
                r.end_time end_time,
                r.process process,
                r.version version,
                r.global_tag,
                rs.status status,
                (SELECT COUNT(s.run_id) FROM streamer s WHERE s.run_id = r.run_id) numStreamers,
                (SELECT COUNT(re.run_id) FROM repacked re WHERE re.run_id = r.run_id) numRepacked,
                (SELECT COUNT(rec.run_id) FROM reconstructed rec WHERE rec.run_id = r.run_id) numReconstructed
                FROM run r, run_status rs
                WHERE rs.id = r.run_status
                ORDER BY """ + str(int(self.orderby) + 1) + " " + order + """
            ) sub WHERE rownum <= :numruns"""

        self.bindParams = {'numruns' : int(self.numruns)}

#        globals.logger.log(self.query)
